from typing import Iterable, Generator, AsyncIterable, AsyncGenerator

import itertools


def grouper(n: int, iterable: Iterable) -> Generator:
    it = iter(iterable)
    while True:
        chunk = tuple(itertools.islice(it, n))
        if not chunk:
            return
        yield chunk


async def aenumerate(async_iterable: AsyncIterable, start: int = 0, /) -> AsyncGenerator:
    i = start
    async for element in async_iterable:
        yield i, element
        i += 1


async def async_grouper(n: int, async_iterable: AsyncIterable) -> AsyncGenerator:
    chunk = []
    async for i, element in aenumerate(async_iterable, 1):
        chunk.append(element)
        if i % n == 0:
            yield tuple(chunk)
            chunk = []
    if chunk:
        yield chunk



import os

from typing import Optional, List, Any

from jara_utils.exceptions import EnvironmentVariableNotFound
from jara_utils.normalization.boolean import str_2_bool
from jara_utils.normalization.integer import parse_int


def get_str(name: str, invalid: Optional[str] = None, required: bool = False) -> Optional[str]:
    if name not in os.environ and required:
        raise EnvironmentVariableNotFound(name)
    return str(os.environ.get(name, '')) or invalid


def get_int(name: str, invalid: Optional[int] = None, required: bool = False) -> Optional[int]:
    if name not in os.environ and required:
        raise EnvironmentVariableNotFound(name)
    return parse_int(os.environ.get(name)) or invalid


def get_bool(name: str, invalid: Optional[bool] = None, required: bool = False) -> Optional[bool]:
    if name not in os.environ and required:
        raise EnvironmentVariableNotFound(name)
    value = str_2_bool(os.environ.get(name))
    return value if value is not None else invalid


def get_list(name: str, separator: str = ',', invalid: Optional[List[Any]] = None, required: bool = False) -> List[str]:
    if name not in os.environ and required:
        raise EnvironmentVariableNotFound(name)

    if invalid is None:
        invalid = []

    value = os.environ.get(name)
    return value.split(separator) if value else invalid



class JaraError(Exception):
    def __init__(self, message: str):
        self.message = message

    def __str__(self):
        return self.message

    def __unicode__(self):  # pragma: no cover
        return self.message


class EnvironmentVariableNotFound(JaraError):
    """Raise this exception when an environment variable is not defined, but should be."""
    def __init__(self, variable_name: str):
        super().__init__(f'Environment variable with name {variable_name} not found!')



from datetime import datetime, date
from typing import Any

import pytz


def date_2_datetime(value: date, naive: bool = False) -> datetime:
    result = datetime.combine(value, datetime.min.time())
    if naive:
        return result
    return result.replace(tzinfo=pytz.UTC)


def str_2_datetime(value: str, fmt: str = '%Y-%m-%d', naive: bool = False, invalid: Any = datetime.min) -> datetime:
    try:
        result = datetime.strptime(value, fmt)
    except ValueError:
        result = invalid
    except TypeError:
        result = invalid

    if not result:
        return invalid

    if naive:
        return result
    return result.replace(tzinfo=pytz.UTC)



class TextColor:
    ENDC = '\033[0m'  # default format

    CYAN = '\033[0;36m'  # debug
    OKBLUE = '\033[94m'  # info
    OKGREEN = '\033[92m'  # success
    YELLOW = '\033[93m'  # warning
    RED = '\033[91m'  # error/fail

    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

    MAGENTA = '\033[0;35m'



import re
from typing import Optional


def str_2_bool(val: Optional[str]) -> Optional[bool]:
    if not val:
        return None
    if val.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif val.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    return None


def is_dunder(attr: str) -> bool:
    if len(attr) < 5:
        return False
    return attr.startswith('__') and attr.endswith('__')


def is_email(value: str) -> bool:
    if not value:
        return False
    user, _, domain = value.partition('@')
    if not all([user, domain]):
        return False

    user_regex = re.compile(
        r"(^[-!#$%&'*+/=?^`{}|~\w]+(\.[-!#$%&'*+/=?^`{}|~\w]+)*$"  # dot-atom
        # quoted-string
        r'|^"([\001-\010\013\014\016-\037!#-\[\]-\177]'
        r'|\\[\001-\011\013\014\016-\177])*"$)', re.IGNORECASE | re.UNICODE)

    domain_regex = re.compile(
        # domain
        r'(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+'
        r'(?:[A-Z]{2,6}|[A-Z0-9-]{2,})$'
        # literal form, ipv4 address (SMTP 4.1.3)
        r'|^\[(25[0-5]|2[0-4]\d|[0-1]?\d?\d)'
        r'(\.(25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3}\]$', re.IGNORECASE | re.UNICODE)

    if not user_regex.match(user):
        return False
    if not domain_regex.match(domain):
        return False
    return True



import re
import sys
from typing import TextIO, Optional, List, Union

from jara_utils.constants import TextColor


class CLIOutput:
    def __init__(self, stream: TextIO = sys.stdout, error_stream: TextIO = sys.stderr):
        self.stream = stream
        self.error_stream = error_stream
        self.hl_pattern = re.compile(r'\[[\w\d\-.,\s<>\':]+\]')

    def _wrap(self, message: str, color: str, highlights: Optional[List[str]] = None) -> str:
        message = f'{color}{message}{TextColor.ENDC}'
        if highlights is None:
            highlights = re.findall(self.hl_pattern, message)
        for hl in highlights:
            message = message.replace(hl, f'{TextColor.BOLD}{hl}{TextColor.ENDC}{color}')
        return message

    def write_stream(self, message: str, crlf: str):
        self.stream.write(message + crlf)

    def write_error(self, message: str, crlf: str):
        self.stream.write(message + crlf)

    def debug(self, text: str, highlights: Optional[List[str]] = None, crlf: str = '\n'):
        self.write_stream(self._wrap(text, TextColor.CYAN, highlights), crlf)

    def info(self, text: str, highlights: Optional[List[str]] = None, crlf: str = '\n'):
        self.write_stream(self._wrap(text, TextColor.OKBLUE, highlights), crlf)

    def success(self, text: str, highlights: Optional[List[str]] = None, crlf: str = '\n'):
        self.write_stream(self._wrap(text, TextColor.OKGREEN, highlights), crlf)

    def warning(self, text: str, highlights: Optional[List[str]] = None, crlf: str = '\n'):
        self.write_stream(self._wrap(text, TextColor.YELLOW, highlights), crlf)

    def error(self, text: Union[str], highlights: Optional[List[str]] = None, crlf: str = '\n'):
        self.write_error(self._wrap(text, TextColor.RED, highlights), crlf)

    def fail(self, text: Union[str], highlights: Optional[List[str]] = None, crlf: str = '\n'):
        self.error(text, highlights, crlf)



import inspect
import logging
import time
from functools import wraps
from typing import Any, Callable, Optional

log = logging.getLogger(__name__)


def timeit(*args, **kwargs) -> Any:
    """A method decorator which can be applied on async and normal methods."""

    def _log_execution(start: float, name: Optional[Any], coro: Callable, coro_args: Any):
        total = time.time() - start
        if total >= 0.5:  # pragma: no cover
            log.warning(f'{name if name else coro.__name__} finished after %.3f seconds' % total)

    def inner(coro: Callable) -> Callable:
        @wraps(coro)
        async def wrapped(*coro_args, **coro_kwargs) -> Any:
            name = kwargs.get('name')
            start = time.time()
            try:
                return await coro(*coro_args, **coro_kwargs)
            finally:
                _log_execution(start, name, coro, coro_args)

        @wraps(coro)
        def timed(*coro_args, **coro_kwargs) -> Any:
            name = kwargs.get('name')
            start = time.time()

            try:
                return coro(*coro_args, **coro_kwargs)
            finally:
                _log_execution(start, name, coro, coro_args)

        if inspect.iscoroutinefunction(coro):
            return wrapped
        return timed

    if args:
        return inner(*args)
    else:
        return inner



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jara_utils', 'jara_utils.decorators', 'jara_utils.normalization']

package_data = \
{'': ['*']}

install_requires = \
['pytz>=2022.7.1,<2023.0.0', 'types-pytz>=2022.7.1.0,<2023.0.0.0']

setup_kwargs = {
    'name': 'jara-utils',
    'version': '1.0.0',
    'description': 'A package with basic stuff.',
    'long_description': '=================\nJara core package\n=================\n\n    Jara means bear in `Sesotho`_.\n\n.. image:: https://img.shields.io/badge/python-3.11.x-blue.svg\n    :alt: Python 3.11.x\n\n.. image:: https://img.shields.io/badge/code_style-flake8-brightgreen.svg\n    :alt: Flake8\n\n.. image:: https://img.shields.io/badge/dependency_manager-poetry-blueviolet.svg\n    :alt: Poetry\n\nWhy?\n~~~~\nSometimes I start a new project and I need to implement again same methods and after create tests for each method. This package will provide common methods like ``str_2_bool`` or other methods check ``README.rst`` for all methods available.\n\nHow to contribute to the package?\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\nClone project locally and then:\n    * Install all dependencies including the test and the dev oanes: ``poetry install -E test -D``;\n    * Do changes in the project;\n    * Create unittests (please make sure  you will keep coverage to 100%);\n    * Run all sanity commands (pytest, flake8, mypy, bandit);\n    * Check if there is any duplicated or dead fixtures by running ``pytest`` with ``--dead-fixtures`` and ``--dup-fixtures``;\n\nNote: Run commands using poetry: ``poetry run <command>``;\n\nWhat you will find in this package?\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\nBasically will contain utils methods to avoid write them in all projects. Some examples:\n\n* decorator to benchmark methods;\n* methods to handle environment variables;\n* some utils methods such as: ``snake_2_camel``, ``str_2_bool``.\n\n\n.. _Sesotho: https://en.wikipedia.org/wiki/Sotho_language\n',
    'author': 'Serban Senciuc',
    'author_email': 'senciuc.serban@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/senciucserban/jara-utils',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9.16,<4.0.0',
}


setup(**setup_kwargs)



import re


def snake_to_camel(s: str, upper: bool = True) -> str:
    """
        Convert from snake_case to CamelCase
        If upper is set, then convert to upper CamelCase, otherwise the first character keeps its case.
    """
    _re_snake_to_camel = re.compile(r'(_)([a-z\d])')
    s = _re_snake_to_camel.sub(lambda m: m.group(2).upper(), s)
    if upper:
        s = s[:1].upper() + s[1:]
    return s


def camel_to_snake(s: str) -> str:
    """
        Convert from CamelCase to snake_case
    """
    _re_camel_to_snake = re.compile(r'([a-z]|[A-Z]+)(?=[A-Z])')
    return _re_camel_to_snake.sub(r'\1_', s).lower()



from datetime import datetime, date
from typing import Optional, Any


def parse_int(value: Any) -> Optional[int]:
    try:
        return int(value)
    except (ValueError, TypeError):
        return None


def age_from(birthday: Optional[datetime]) -> Optional[int]:
    if not birthday or birthday == datetime.min:
        return None
    today = date.today()
    return today.year - birthday.year - ((today.month, today.day) < (birthday.month, birthday.day))



from .boolean import str_2_bool, is_dunder, is_email
from .datetime import date_2_datetime, str_2_datetime
from .integer import parse_int, age_from
from .iterable import grouper, async_grouper, aenumerate
from .string import snake_to_camel, camel_to_snake


__all__ = (
    'str_2_bool', 'is_dunder', 'is_email',
    'date_2_datetime', 'str_2_datetime',
    'parse_int', 'age_from',
    'grouper', 'async_grouper', 'aenumerate',
    'snake_to_camel', 'camel_to_snake',
)
